@@ -37,6 +37,8 @@
 #include <osmocom/core/gsmtap_util.h>
 #include <osmocom/core/utils.h>
 
+#include <osmocom/codec/codec.h>
+
 #include <osmocom/trau/osmo_ortp.h>
 
 #include <osmo-bts/logging.h>
@@ -52,39 +54,16 @@
 #include <osmo-bts/bts_model.h>
 #include <osmo-bts/handover.h>
 #include <osmo-bts/msg_utils.h>
+#include <osmo-bts/rtp_input_preen.h>
 #include <osmo-bts/pcuif_proto.h>
 #include <osmo-bts/cbch.h>
-
-
-#define CB_FCCH		-1
-#define CB_SCH		-2
-#define CB_BCCH		-3
-#define CB_IDLE		-4
-
-/* according to TS 05.02 Clause 7 Table 3 of 9 an Figure 8a */
-static const int ccch_block_table[51] = {
-	CB_FCCH, CB_SCH,/* 0..1 */
-	CB_BCCH, CB_BCCH, CB_BCCH, CB_BCCH,	/* 2..5: BCCH */
-	0, 0, 0, 0,	/* 6..9: B0 */
-	CB_FCCH, CB_SCH,/* 10..11 */
-	1, 1, 1, 1,	/* 12..15: B1 */
-	2, 2, 2, 2,	/* 16..19: B2 */
-	CB_FCCH, CB_SCH,/* 20..21 */
-	3, 3, 3, 3,	/* 22..25: B3 */
-	4, 4, 4, 4,	/* 26..29: B4 */
-	CB_FCCH, CB_SCH,/* 30..31 */
-	5, 5, 5, 5,	/* 32..35: B5 */
-	6, 6, 6, 6,	/* 36..39: B6 */
-	CB_FCCH, CB_SCH,/* 40..41 */
-	7, 7, 7, 7, 	/* 42..45: B7 */
-	8, 8, 8, 8, 	/* 46..49: B8 */
-	-4		/* 50: Idle */
-};
+#include <osmo-bts/asci.h>
+#include <osmo-bts/csd_v110.h>
 
 /* determine the CCCH block number based on the frame number */
 unsigned int l1sap_fn2ccch_block(uint32_t fn)
 {
-	int rc = ccch_block_table[fn%51];
+	int rc = gsm0502_fn2ccch_block(fn);
 	/* if FN is negative, we were called for something that's not CCCH! */
 	OSMO_ASSERT(rc >= 0);
 	return rc;
@@ -615,6 +594,9 @@ static void l1sap_interf_meas_report(struct gsm_bts *bts)
 		return;
 
 	llist_for_each_entry(trx, &bts->trx_list, list) {
+		if (trx->mo.nm_state.operational != NM_OPSTATE_ENABLED ||
+		    trx->bb_transc.mo.nm_state.operational != NM_OPSTATE_ENABLED)
+			continue;
 		/* Calculate the average of all received samples */
 		l1sap_interf_meas_calc_avg(trx);
 		/* Report to the BSC over the A-bis/RSL */
@@ -697,6 +679,10 @@ static void process_l1sap_meas_data(struct gsm_lchan *lchan,
 	uint32_t fn;
 	const char *ind_name;
 
+	/* Do not process measurement reports from non-active VGCS calls. */
+	if (rsl_chan_rt_is_asci(lchan->rsl_chan_rt) && lchan->asci.talker_active != VGCS_TALKER_ACTIVE)
+		return;
+
 	switch (ind_type) {
 	case PRIM_MPH_INFO:
 		/* (legacy way, see also OS#2977) */
@@ -748,10 +734,9 @@ static void process_l1sap_meas_data(struct gsm_lchan *lchan,
 		OSMO_ASSERT(false);
 	}
 
-	DEBUGPFN(DL1P, fn,
-		 "%s %s meas ind, ta_offs_256bits=%d, ber10k=%d, inv_rssi=%u, C/I=%d cB\n",
-		 gsm_lchan_name(lchan), ind_name, ulm.ta_offs_256bits,
-		 ulm.ber10k, ulm.inv_rssi, ulm.ci_cb);
+	LOGPLCFN(lchan, fn, DL1P, LOGL_DEBUG,
+		 "%s meas ind, ta_offs_256bits=%d, ber10k=%d, inv_rssi=%u, C/I=%d cB\n", ind_name,
+		 ulm.ta_offs_256bits, ulm.ber10k, ulm.inv_rssi, ulm.ci_cb);
 
 	/* we assume that symbol period is 1 bit: */
 	set_ms_to_data(lchan, ulm.ta_offs_256bits / 256, true);
@@ -903,14 +888,13 @@ static int lchan_pdtch_ph_rts_ind_loop(struct gsm_lchan *lchan,
 	/* de-queue response message (loopback) */
 	loop_msg = msgb_dequeue_count(&lchan->dl_tch_queue, &lchan->dl_tch_queue_len);
 	if (!loop_msg) {
-		LOGPGT(DL1P, LOGL_NOTICE, tm, "%s: no looped PDTCH message, sending empty\n",
-		     gsm_lchan_name(lchan));
+		LOGPLCGT(lchan, tm, DL1P, LOGL_NOTICE, "no looped PDTCH message, sending empty\n");
 		/* empty downlink message */
 		p = msgb_put(msg, GSM_MACBLOCK_LEN);
 		memset(p, 0, GSM_MACBLOCK_LEN);
 	} else {
-		LOGPGT(DL1P, LOGL_NOTICE, tm, "%s: looped PDTCH message of %u bytes\n",
-		     gsm_lchan_name(lchan), msgb_l2len(loop_msg));
+		LOGPLCGT(lchan, tm, DL1P, LOGL_NOTICE, "looped PDTCH message of %u bytes\n",
+			 msgb_l2len(loop_msg));
 		/* copy over data from queued response message */
 		p = msgb_put(msg, msgb_l2len(loop_msg));
 		memcpy(p, msgb_l2(loop_msg), msgb_l2len(loop_msg));
@@ -919,15 +903,32 @@ static int lchan_pdtch_ph_rts_ind_loop(struct gsm_lchan *lchan,
 	return 0;
 }
 
-/* Check if given CCCH frame number is for a PCH or for an AGCH (this function is
+/* Check if given CCCH frame number is for a NCH, PCH or for an AGCH (this function is
  * only used internally, it is public to call it from unit-tests) */
-int is_ccch_for_agch(struct gsm_bts_trx *trx, uint32_t fn) {
+enum ccch_msgt get_ccch_msgt(struct gsm_bts_trx *trx, uint32_t fn)
+{
+	uint8_t block, first_block, num_blocks;
+	int rc;
+
+	block = l1sap_fn2ccch_block(fn);
+
+	/* If there is an NCH, check if the block number matches. It has priority over PCH/AGCH. */
+	if (trx->bts->asci.pos_nch >= 0) {
+		rc = osmo_gsm48_si1ro_nch_pos_decode(trx->bts->asci.pos_nch, &num_blocks, &first_block);
+		if (rc >= 0 && block >= first_block && block < first_block + num_blocks)
+			return CCCH_MSGT_NCH;
+	}
+
 	/* Note: The number of available access grant channels is set by the
 	 * parameter BS_AG_BLKS_RES via system information type 3. This SI is
 	 * transferred to osmo-bts via RSL */
-        return l1sap_fn2ccch_block(fn) < num_agch(trx, "PH-RTS-IND");
+	if (l1sap_fn2ccch_block(fn) < num_agch(trx, "PH-RTS-IND"))
+		return CCCH_MSGT_AGCH;
+
+	return CCCH_MSGT_PCH;
 }
 
+
 /* return the measured average of frame numbers that the RTS clock is running in advance */
 int32_t bts_get_avg_fn_advance(const struct gsm_bts *bts)
 {
@@ -1065,7 +1066,6 @@ static int l1sap_ph_rts_ind(struct gsm_bts_trx *trx,
 	struct msgb *pp_msg;
 	bool dtxd_facch = false;
 	int rc;
-	int is_ag_res;
 
 	chan_nr = rts_ind->chan_nr;
 	link_id = rts_ind->link_id;
@@ -1125,9 +1125,7 @@ static int l1sap_ph_rts_ind(struct gsm_bts_trx *trx,
 			return 0;
 		}
 		if (lchan->pending_rel_ind_msg) {
-			LOGPGT(DRSL, LOGL_INFO, &g_time,
-			       "%s Forward RLL RELease INDication to the BSC\n",
-			       gsm_lchan_name(lchan));
+			LOGPLCGT(lchan, &g_time, DRSL, LOGL_INFO, "Forward RLL RELease INDication to the BSC\n");
 			abis_bts_rsl_sendmsg(lchan->pending_rel_ind_msg);
 			lchan->pending_rel_ind_msg = NULL;
 		}
@@ -1162,6 +1160,7 @@ static int l1sap_ph_rts_ind(struct gsm_bts_trx *trx,
 				pp_msg = lapdm_phsap_dequeue_msg_facch(lchan, le, fn);
 			else
 				pp_msg = lapdm_phsap_dequeue_msg(le);
+			lchan->tch.dtx_fr_hr_efr.dl_facch_stealing = (pp_msg != NULL);
 		}
 		if (!pp_msg) {
 			if (L1SAP_IS_LINK_SACCH(link_id)) {
@@ -1172,6 +1171,10 @@ static int l1sap_ph_rts_ind(struct gsm_bts_trx *trx,
 					memcpy(p + 2, si, GSM_MACBLOCK_LEN - 2);
 				} else
 					memcpy(p + 2, fill_frame, GSM_MACBLOCK_LEN - 2);
+			} else if (vgcs_is_uplink_free(lchan)) {
+				/* If UPLINK FREE message is stored, send it with every DCCH frame. */
+				p = msgb_put(msg, GSM_MACBLOCK_LEN);
+				vgcs_uplink_free_get(lchan, p);
 			} else if (L1SAP_IS_CHAN_SDCCH4(chan_nr) || L1SAP_IS_CHAN_SDCCH8(chan_nr) ||
 				   (lchan->rsl_cmode == RSL_CMOD_SPD_SIGN && !lchan->ts->trx->bts->dtxd)) {
 				/*
@@ -1196,13 +1199,20 @@ static int l1sap_ph_rts_ind(struct gsm_bts_trx *trx,
 				check_for_ciph_cmd(pp_msg, lchan, chan_nr);
 				if (dtxd_facch)
 					dtx_dispatch(lchan, E_FACCH);
+				if (rsl_chan_rt_is_vgcs(lchan->rsl_chan_rt)) {
+					/* Check for UPLINK FREE message and store. */
+					if (pp_msg->data[0] == GSM48_MT_RR_SH_UL_FREE << 2)
+						vgcs_uplink_free_set(lchan, pp_msg->data);
+					/* Keep UPLINK FREE message when sending short header messages. */
+					else if ((pp_msg->data[0] & 0x03) != 0x00)
+						vgcs_uplink_free_reset(lchan);
+				}
 			}
 			msgb_free(pp_msg);
 		}
 	} else if (L1SAP_IS_CHAN_AGCH_PCH(chan_nr)) {
 		p = msgb_put(msg, GSM_MACBLOCK_LEN);
-		is_ag_res = is_ccch_for_agch(trx, fn);
-		rc = bts_ccch_copy_msg(trx->bts, p, &g_time, is_ag_res);
+		rc = bts_ccch_copy_msg(trx->bts, p, &g_time, get_ccch_msgt(trx, fn));
 		if (rc <= 0)
 			memcpy(p, fill_frame, GSM_MACBLOCK_LEN);
 	}
@@ -1215,45 +1225,267 @@ static int l1sap_ph_rts_ind(struct gsm_bts_trx *trx,
 	return 1;
 }
 
-static bool rtppayload_is_octet_aligned(const uint8_t *rtp_pl, uint8_t payload_len)
+/* The following static functions are helpers for l1sap_tch_rts_ind(),
+ * used only for FR/HR/EFR speech modes.  For these speech TCH modes,
+ * if our incoming RTP stream includes SID frames, we need to apply
+ * the following transformations to the downlink frame stream we actually
+ * transmit:
+ *
+ * - We need to cache the last received SID and retransmit it in
+ *   SACCH-aligned frame positions, or if the SACCH-aligned frame
+ *   position was stolen by FACCH, then right after that FACCH.
+ *
+ * - That cached SID needs to be aged and expired, in accord with
+ *   TS 28.062 section C.3.2.1.1 paragraph 5 - the paragraph concerning
+ *   expiration of cached downlink SID.
+ *
+ * - In all other frame positions, extraneous SID frames need to be
+ *   dropped - we send an empty payload to the BTS model, causing it
+ *   to transmit an induced BFI condition on the air (fake DTXd),
+ *   or perhaps real DTXd (actually turning off Tx) in the future.
+ */
+
+/*! \brief Check if the given FN of TCH-RTS-IND corresponds to a mandatory
+ *  SID position for non-AMR codecs that follow SACCH alignment.
+ *  \param[in] lchan Logical channel on which we check scheduling
+ *  \param[in] fn Frame Number for which we check scheduling
+ *  \returns true if this FN is a mandatory SID position, false otherwise
+ */
+static inline bool fr_hr_efr_sid_position(struct gsm_lchan *lchan, uint32_t fn)
 {
-	/*
-	 * Logic: If 1st bit padding is not zero, packet is either:
-	 * - bandwidth-efficient AMR payload.
-	 * - malformed packet.
-	 * However, Bandwidth-efficient AMR 4,75 frame last in payload(F=0, FT=0)
-	 * with 4th,5ht,6th AMR payload to 0 matches padding==0.
-	 * Furthermore, both AMR 4,75 bw-efficient and octet alignment are 14 bytes long (AMR 4,75 encodes 95b):
-	 * bw-efficient: 95b, + 4b hdr + 6b ToC = 105b, + padding = 112b = 14B.
-	 * octet-aligned: 1B hdr + 1B ToC + 95b = 111b, + padding = 112b = 14B.
-	 * We cannot use other fields to match since they are inside the AMR
-	 * payload bits which are unknown.
-	 * As a result, this function may return false positive (true) for some AMR
-	 * 4,75 AMR frames, but given the length, CMR and FT read is the same as a
-	 * consequence, the damage in here is harmless other than being unable to
-	 * decode the audio at the other side.
-	 */
-	#define AMR_PADDING1(rtp_pl) (rtp_pl[0] & 0x0f)
-	#define AMR_PADDING2(rtp_pl) (rtp_pl[1] & 0x03)
-
-	if (payload_len < 2 || AMR_PADDING1(rtp_pl) || AMR_PADDING2(rtp_pl))
+	/* See GSM 05.08 section 8.3 for frame numbers - but we are
+	 * specifically looking for FNs corresponding to the beginning
+	 * of the complete SID frame to be transmitted, rather than all FNs
+	 * where we have to put out a non-suppressed burst. */
+	switch (lchan->type) {
+	case GSM_LCHAN_TCH_F:
+		return fn % 104 == 52;
+	case GSM_LCHAN_TCH_H:
+		switch (lchan->nr) {
+		case 0:
+			return fn % 104 == 0 || fn % 104 == 52;
+		case 1:
+			return fn % 104 == 14 || fn % 104 == 66;
+		default:
+			return false;
+		}
+	default:
 		return false;
+	}
+}
 
+/*! \brief This helper function implements DTXd input processing for FR/HR/EFR:
+ *  we got an RTP input frame, now we need to check if it is SID or not,
+ *  and update our DL SID reshaper state accordingly.
+ *  \param[in] lchan Logical channel structure of the TCH we work with
+ *  \param[in] resp_msg The input frame from RTP
+ *  \param[out] sid_result Output flag indicating if the received frame is SID
+ *  \returns true if the frame in resp_msg is good, false otherwise
+ */
+static bool fr_hr_efr_dtxd_input(struct gsm_lchan *lchan, struct msgb *resp_msg,
+				 bool *sid_result)
+{
+	enum osmo_gsm631_sid_class sidc;
+	bool is_sid;
+
+	switch (lchan->tch_mode) {
+	case GSM48_CMODE_SPEECH_V1:
+		if (lchan->type == GSM_LCHAN_TCH_F) {
+			sidc = osmo_fr_sid_classify(msgb_l2(resp_msg));
+			switch (sidc) {
+			case OSMO_GSM631_SID_CLASS_SPEECH:
+				is_sid = false;
+				break;
+			case OSMO_GSM631_SID_CLASS_INVALID:
+				/* TS 28.062 section C.3.2.1.1: invalid SIDs
+				 * from call leg A UL are treated like BFIs.
+				 * Drop this frame and act as if we got nothing
+				 * at all from RTP for this FN. */
+				return false;
+			case OSMO_GSM631_SID_CLASS_VALID:
+				is_sid = true;
+				/* The SID code word may have a one bit error -
+				 * rejuvenate it. */
+				osmo_fr_sid_reset(msgb_l2(resp_msg));
+				break;
+			default:
+				/* SID classification per GSM 06.31 section
+				 * 6.1.1 has only 3 possible outcomes. */
+				OSMO_ASSERT(0);
+			}
+		} else {
+			/* The same kind of classification as we do in
+			 * osmo_{fr,efr}_sid_classify() is impossible for HR:
+			 * the equivalent ternary SID classification per
+			 * GSM 06.41 can only be done in the UL-handling BTS,
+			 * directly coupled to the GSM 05.03 channel decoder,
+			 * and cannot be reconstructed downstream from frame
+			 * payload bits.  The only kind of SID we can detect
+			 * here is the perfect, error-free kind.  However,
+			 * if we received RFC 5993 payload and the sender
+			 * told us it is valid SID, honor that indication
+			 * and rejuvenate the SID codeword. */
+			if (rtpmsg_is_rfc5993_sid(resp_msg)) {
+				is_sid = true;
+				osmo_hr_sid_reset(msgb_l2(resp_msg));
+			} else {
+				is_sid = osmo_hr_check_sid(msgb_l2(resp_msg),
+							   msgb_l2len(resp_msg));
+			}
+		}
+		break;
+	case GSM48_CMODE_SPEECH_EFR:
+		/* same logic as for FRv1 */
+		sidc = osmo_efr_sid_classify(msgb_l2(resp_msg));
+		switch (sidc) {
+		case OSMO_GSM631_SID_CLASS_SPEECH:
+			is_sid = false;
+			break;
+		case OSMO_GSM631_SID_CLASS_INVALID:
+			return false;
+		case OSMO_GSM631_SID_CLASS_VALID:
+			is_sid = true;
+			osmo_efr_sid_reset(msgb_l2(resp_msg));
+			break;
+		default:
+			OSMO_ASSERT(0);
+		}
+		break;
+	default:
+		/* This static function should never be called except for
+		 * V1 and EFR speech modes. */
+		OSMO_ASSERT(0);
+	}
+	*sid_result = is_sid;
+	lchan->tch.dtx_fr_hr_efr.last_rtp_input_was_sid = is_sid;
+	if (is_sid) {
+		uint8_t copy_len;
+
+		copy_len = OSMO_MIN(msgb_l2len(resp_msg),
+				ARRAY_SIZE(lchan->tch.dtx_fr_hr_efr.last_sid));
+		memcpy(lchan->tch.dtx_fr_hr_efr.last_sid,
+			msgb_l2(resp_msg), copy_len);
+		lchan->tch.dtx_fr_hr_efr.last_sid_len = copy_len;
+		lchan->tch.dtx_fr_hr_efr.last_sid_age = 0;
+	} else {
+		/* We got a speech frame, not SID - therefore, the state flag
+		 * of "we already transmitted a SID" needs to be cleared,
+		 * so that the very first SID that follows this talkspurt
+		 * will get transmitted right away, without waiting for
+		 * the next mandatory SID position. */
+		lchan->tch.dtx_fr_hr_efr.dl_sid_transmitted = false;
+	}
 	return true;
 }
 
-static bool rtppayload_is_valid(struct gsm_lchan *lchan, struct msgb *resp_msg)
+/*! \brief This helper function implements DTXd output processing for FR/HR/EFR:
+ *  here we update our state to deal with cached SID aging, mandatory-Tx
+ *  frame positions and FACCH stealing, and we make the desired output
+ *  transformations of either regurgitating a cached SID or vice-versa,
+ *  dropping a SID we received from RTP.
+ *  \param[in] lchan Logical channel structure of the TCH we work with
+ *  \param[in] fn Frame Number for which we are preparing DL
+ *  \param[in] current_input_is_sid Self-explanatory
+ *  \param[in] resp_msg_p Pointer to l1sap_tch_rts_ind() internal variable
+ *  \param[in] resp_l1sap_p ditto
+ *  \param[in] empty_l1sap_p ditto
+ *
+ *  This function gets called with pointers to l1sap_tch_rts_ind() internal
+ *  variables and cannot be properly understood on its own, without
+ *  understanding the parent function first.  This situation is unfortunate,
+ *  but it was the only way to factor the present logic out of
+ *  l1sap_tch_rts_ind() main body.
+ */
+static void fr_hr_efr_dtxd_output(struct gsm_lchan *lchan, uint32_t fn,
+				  bool current_input_is_sid,
+				  struct msgb **resp_msg_p,
+				  struct osmo_phsap_prim **resp_l1sap_p,
+				  struct osmo_phsap_prim *empty_l1sap_p)
 {
-	/* Avoid sending bw-efficient AMR to lower layers, most bts models
-	 * don't support it. */
-	if (lchan->tch_mode == GSM48_CMODE_SPEECH_AMR &&
-		!rtppayload_is_octet_aligned(resp_msg->data, resp_msg->len)) {
-		LOGPLCHAN(lchan, DL1P, LOGL_NOTICE,
-			  "RTP->L1: Dropping unexpected AMR encoding (bw-efficient?) %s\n",
-			  osmo_hexdump(resp_msg->data, resp_msg->len));
-		return false;
+	struct msgb *resp_msg = *resp_msg_p;
+	bool sid_in_hand = current_input_is_sid;
+
+	/* Are we at a mandatory-Tx frame position? If so, clear the state flag
+	 * of "we already transmitted a SID in this window" - as of right now,
+	 * a SID has _not_ been transmitted in the present window yet, and if
+	 * we are in a DTX pause, we do need to transmit a SID update as soon
+	 * as we are able to, FACCH permitting. */
+	if (fr_hr_efr_sid_position(lchan, fn))
+		lchan->tch.dtx_fr_hr_efr.dl_sid_transmitted = false;
+
+	/* The next stanza implements logic that was originally meant to reside
+	 * in the TFO block in TRAUs: if the source feeding us RTP is in a DTXu
+	 * pause (resp_msg == NULL, meaning we got nothing from RTP) *and* the
+	 * most recent RTP input we did get was a SID, and that SID hasn't
+	 * expired, then we need to replicate that SID. */
+	if (!resp_msg && lchan->tch.dtx_fr_hr_efr.last_rtp_input_was_sid &&
+	    lchan->tch.dtx_fr_hr_efr.last_sid_age < 47) {
+		/* Whatever we do further below, any time another 20 ms window
+		 * has passed since the last SID was received in RTP, we have
+		 * to age that cached SID. */
+		lchan->tch.dtx_fr_hr_efr.last_sid_age++;
+
+		/* The following conditional checking dl_sid_transmitted flag
+		 * is peculiar to our sans-E1 architecture.  In the original
+		 * T1/E1 Abis architecture the TFO-enabled TRAU would repeat
+		 * the cached SID from call leg A into *every* 20 ms frame in
+		 * its Abis output, and then the BTS would select which ones
+		 * it should transmit per the rules of GSM 06.31/06.81 section
+		 * 5.1.2.  But in our architecture it would be silly and
+		 * wasteful to allocate and fill an msgb for this cached SID
+		 * only to toss it later in the same function - hence the
+		 * logic of section 5.1.2 is absorbed into the decision right
+		 * here to proceed with cached SID regurgitation or not,
+		 * in the form of the following conditional. */
+		if (!lchan->tch.dtx_fr_hr_efr.dl_sid_transmitted)
+			resp_msg = l1sap_msgb_alloc(lchan->tch.dtx_fr_hr_efr.last_sid_len);
+		if (resp_msg) {
+			resp_msg->l2h = msgb_put(resp_msg,
+						lchan->tch.dtx_fr_hr_efr.last_sid_len);
+			memcpy(resp_msg->l2h, lchan->tch.dtx_fr_hr_efr.last_sid,
+				lchan->tch.dtx_fr_hr_efr.last_sid_len);
+			*resp_msg_p = resp_msg;
+			*resp_l1sap_p = msgb_l1sap_prim(resp_msg);
+			sid_in_hand = true;
+		}
+	} else if (resp_msg && sid_in_hand) {
+		/* This "else if" leg implements the logic of section 5.1.2
+		 * for cases when a SID is already present in the RTP input,
+		 * as indicated by (resp_msg != NULL) and sid_in_hand being
+		 * true.  Because we are in the "else" clause of the big "if"
+		 * above, this path executes only when the SID has come from
+		 * RTP in _this_ frame, rather than regurgitated from cache.
+		 * But be it fresh or cached, the rules of section 5.1.2 still
+		 * apply, and if we've already transmitted a SID in the current
+		 * window, then we need to suppress further SIDs and send
+		 * an empty payload to the BTS model, causing the latter
+		 * to transmit an induced BFI condition on the air.  This
+		 * strange-seeming behavior is needed so that the spec-compliant
+		 * Rx DTX handler in the MS will produce the expected output,
+		 * same as if the GSM network were the old-fashioned kind with
+		 * Abis TRAUs and TFO. */
+		if (lchan->tch.dtx_fr_hr_efr.dl_sid_transmitted) {
+			msgb_free(resp_msg);
+			resp_msg = NULL;
+			*resp_msg_p = NULL;
+			*resp_l1sap_p = empty_l1sap_p;
+		}
 	}
-	return true;
+
+	/* The following conditional answers this question: are we actually
+	 * transmitting a SID frame on the air right now, at this frame number?
+	 * If we are certain the BTS model is going to transmit this SID,
+	 * we set the state flag so we won't be transmitting any more SIDs
+	 * until we either hit the next mandatory-Tx position or get another
+	 * little talkspurt followed by new SID.  The check for FACCH stealing
+	 * is included because if the BTS model is going to transmit FACCH in
+	 * the current FN, then we are not actually transmitting SID right now,
+	 * and we still need to transmit a SID ASAP, as soon as the TCH becomes
+	 * becomes free from FACCH activity.  GSM 06.31/06.81 section 5.1.2
+	 * does mention that if the mandatory-Tx frame position is taken up
+	 * by FACCH, then we need to send SID in the following frame. */
+	if (resp_msg && sid_in_hand && !lchan->tch.dtx_fr_hr_efr.dl_facch_stealing)
+		lchan->tch.dtx_fr_hr_efr.dl_sid_transmitted = true;
 }
 
 /* TCH-RTS-IND prim received from bts model */
@@ -1266,18 +1498,19 @@ static int l1sap_tch_rts_ind(struct gsm_bts_trx *trx,
 	struct gsm_lchan *lchan;
 	uint8_t chan_nr, marker = 0;
 	uint32_t fn;
+	bool is_fr_hr_efr_sid = false;
 
 	chan_nr = rts_ind->chan_nr;
 	fn = rts_ind->fn;
 
 	gsm_fn2gsmtime(&g_time, fn);
 
-	DEBUGPGT(DL1P, &g_time, "Rx TCH-RTS.ind chan_nr=%s\n", rsl_chan_nr_str(chan_nr));
-
 	lchan = get_active_lchan_by_chan_nr(trx, chan_nr);
 	if (!lchan) {
 		LOGPGT(DL1P, LOGL_ERROR, &g_time, "No lchan for PH-RTS.ind (chan_nr=%s)\n", rsl_chan_nr_str(chan_nr));
 		return 0;
+	} else {
+		LOGPLCGT(lchan, &g_time, DL1P, LOGL_DEBUG, "Rx TCH-RTS.ind\n");
 	}
 
 	if (!lchan->loopback && lchan->abis_ip.rtp_socket) {
@@ -1293,11 +1526,7 @@ static int l1sap_tch_rts_ind(struct gsm_bts_trx *trx,
 	/* get a msgb from the dl_tx_queue */
 	resp_msg = msgb_dequeue_count(&lchan->dl_tch_queue, &lchan->dl_tch_queue_len);
 	if (!resp_msg) {
-		DEBUGPGT(DL1P, &g_time, "%s DL TCH Tx queue underrun\n", gsm_lchan_name(lchan));
-		resp_l1sap = &empty_l1sap;
-	} else if (!rtppayload_is_valid(lchan, resp_msg)) {
-		msgb_free(resp_msg);
-		resp_msg = NULL;
+		LOGPLCGT(lchan, &g_time, DL1P, LOGL_DEBUG, "DL TCH Tx queue underrun\n");
 		resp_l1sap = &empty_l1sap;
 	} else {
 		/* Obtain RTP header Marker bit from control buffer */
@@ -1307,6 +1536,27 @@ static int l1sap_tch_rts_ind(struct gsm_bts_trx *trx,
 		msgb_push(resp_msg, sizeof(*resp_l1sap));
 		resp_msg->l1h = resp_msg->data;
 		resp_l1sap = msgb_l1sap_prim(resp_msg);
+
+		/* FR/HR/EFR SID or non-SID input handling */
+		if (lchan->tch_mode == GSM48_CMODE_SPEECH_V1 ||
+		    lchan->tch_mode == GSM48_CMODE_SPEECH_EFR) {
+			bool drop;
+
+			drop = !fr_hr_efr_dtxd_input(lchan, resp_msg,
+						     &is_fr_hr_efr_sid);
+			if (OSMO_UNLIKELY(drop)) {
+				msgb_free(resp_msg);
+				resp_msg = NULL;
+				resp_l1sap = &empty_l1sap;
+			}
+		}
+	}
+
+	/* FR/HR/EFR DTXd output stage */
+	if (lchan->tch_mode == GSM48_CMODE_SPEECH_V1 ||
+	    lchan->tch_mode == GSM48_CMODE_SPEECH_EFR) {
+		fr_hr_efr_dtxd_output(lchan, fn, is_fr_hr_efr_sid, &resp_msg,
+					&resp_l1sap, &empty_l1sap);
 	}
 
 	memset(resp_l1sap, 0, sizeof(*resp_l1sap));
@@ -1316,13 +1566,21 @@ static int l1sap_tch_rts_ind(struct gsm_bts_trx *trx,
 	resp_l1sap->u.tch.fn = fn;
 	resp_l1sap->u.tch.marker = marker;
 
-	DEBUGPGT(DL1P, &g_time, "Tx TCH.req chan_nr=%s\n", rsl_chan_nr_str(chan_nr));
+	LOGPLCGT(lchan, &g_time, DL1P, LOGL_DEBUG, "Tx TCH.req\n");
 
 	l1sap_down(trx, resp_l1sap);
 
 	return 0;
 }
 
+/* Reset link timeout to current value. */
+void radio_link_timeout_reset(struct gsm_lchan *lchan)
+{
+	struct gsm_bts *bts = lchan->ts->trx->bts;
+
+	lchan->s = bts->radio_link_timeout.current;
+}
+
 /* process radio link timeout counter S. Follows TS 05.08 Section 5.2
  * "MS Procedure" as the "BSS Procedure [...] shall be determined by the
  * network operator." */
@@ -1330,6 +1588,11 @@ static void radio_link_timeout(struct gsm_lchan *lchan, bool bad_frame)
 {
 	struct gsm_bts *bts = lchan->ts->trx->bts;
 
+	/* Bypass radio link timeout on VGCS/VBS channels: There is no
+	 * uplink SACCH on these when talker is not active. */
+	if (rsl_chan_rt_is_asci(lchan->rsl_chan_rt) && lchan->asci.talker_active != VGCS_TALKER_ACTIVE)
+		return;
+
 	/* Bypass radio link timeout if set to -1 */
 	if (bts->radio_link_timeout.current < 0)
 		return;
@@ -1529,8 +1792,7 @@ static int l1sap_ph_data_ind(struct gsm_bts_trx *trx,
 
 		/* Radio Link Timeout counter */
 		if (len == 0) {
-			LOGPGT(DL1P, LOGL_INFO, &g_time, "%s Lost SACCH block\n",
-			       gsm_lchan_name(lchan));
+			LOGPLCGT(lchan, &g_time, DL1P, LOGL_INFO, "Lost SACCH block\n");
 			radio_link_timeout(lchan, true);
 		} else {
 			radio_link_timeout(lchan, false);
@@ -1548,6 +1810,10 @@ static int l1sap_ph_data_ind(struct gsm_bts_trx *trx,
 	if (lchan->ho.active == HANDOVER_WAIT_FRAME)
 		handover_frame(lchan);
 
+	/* report first valid received frame to VGCS talker process */
+	if (rsl_chan_rt_is_asci(lchan->rsl_chan_rt) && lchan->asci.talker_active == VGCS_TALKER_WAIT_FRAME)
+		vgcs_talker_frame(lchan);
+
 	if (L1SAP_IS_LINK_SACCH(link_id))
 		le = &lchan->lapdm_ch.lapdm_acch;
 	else
@@ -1564,6 +1830,102 @@ static int l1sap_ph_data_ind(struct gsm_bts_trx *trx,
 	return 1;
 }
 
+static void send_ul_rtp_packet_data(struct gsm_lchan *lchan, uint32_t fn,
+				    const uint8_t *data, uint16_t data_len)
+{
+	uint8_t rtp_pl[RFC4040_RTP_PLEN];
+	int rc;
+
+	rc = csd_v110_rtp_encode(lchan, &rtp_pl[0], data, data_len);
+	if (rc < 0)
+		return;
+
+	osmo_rtp_send_frame_ext(lchan->abis_ip.rtp_socket,
+				&rtp_pl[0], sizeof(rtp_pl),
+				fn_ms_adj(fn, lchan),
+				lchan->rtp_tx_marker);
+	/* Only clear the marker bit once we have sent a RTP packet with it */
+	lchan->rtp_tx_marker = false;
+}
+
+/* a helper function for the logic in l1sap_tch_ind() */
+static void send_ul_rtp_packet_speech(struct gsm_lchan *lchan, uint32_t fn,
+				      const uint8_t *rtp_pl, uint16_t rtp_pl_len)
+{
+	if (lchan->abis_ip.osmux.use) {
+		lchan_osmux_send_frame(lchan, rtp_pl, rtp_pl_len,
+				       fn_ms_adj(fn, lchan), lchan->rtp_tx_marker);
+	} else if (lchan->abis_ip.rtp_socket) {
+		osmo_rtp_send_frame_ext(lchan->abis_ip.rtp_socket,
+			rtp_pl, rtp_pl_len, fn_ms_adj(fn, lchan), lchan->rtp_tx_marker);
+	}
+	/* Only clear the marker bit once we have sent a RTP packet with it */
+	lchan->rtp_tx_marker = false;
+}
+
+/* a helper function for emitting HR1 UL in RFC 5993 format */
+static void send_rtp_rfc5993(struct gsm_lchan *lchan, uint32_t fn,
+			     struct msgb *msg)
+{
+	uint8_t toc;
+
+	OSMO_ASSERT(msg->len == GSM_HR_BYTES);
+	/* FIXME: implement proper SID classification per GSM 06.41 section
+	 * 6.1.1; see OS#6036.  Until then, detect error-free SID frames
+	 * using our existing osmo_hr_check_sid() function. */
+	if (osmo_hr_check_sid(msg->data, msg->len))
+		toc = 0x20;
+	else
+		toc = 0x00;
+	msgb_push_u8(msg, toc);
+	send_ul_rtp_packet_speech(lchan, fn, msg->data, msg->len);
+}
+
+/* A helper function for l1sap_tch_ind(): handling BFI
+ *
+ * Please note that we pass the msgb to this function, even though it is
+ * currently not used.  This msgb passing is a provision for adding
+ * support for TRAU-UL-like RTP payload formats like TW-TS-001 that allow
+ * indicating BFI along with deemed-bad frame data bits, just like
+ * GSM 08.60 and 08.61 TRAU-UL frames.
+ */
+static void tch_ul_bfi_handler(struct gsm_lchan *lchan,
+			       const struct gsm_time *g_time, struct msgb *msg)
+{
+	uint32_t fn = g_time->fn;
+	uint8_t ecu_out[GSM_FR_BYTES];
+	int rc;
+
+	/* Are we applying an ECU to this uplink, and are we in a state
+	 * (not DTX pause) where we emit ECU output? */
+	if (lchan->ecu_state && !osmo_ecu_is_dtx_pause(lchan->ecu_state)) {
+		rc = osmo_ecu_frame_out(lchan->ecu_state, ecu_out);
+		/* did it actually give us some output? */
+		if (rc > 0) {
+			/* yes, send it out in RTP */
+			send_ul_rtp_packet_speech(lchan, fn, ecu_out, rc);
+			return;
+		}
+	}
+
+	/* Are we in rtp continuous-streaming special mode? If so, send out
+	 * a BFI packet as zero-length RTP payload. */
+	if (lchan->ts->trx->bts->rtp_nogaps_mode) {
+		send_ul_rtp_packet_speech(lchan, fn, NULL, 0);
+		return;
+	}
+
+	/* Most classic form of BFI handling: generate an intentional gap
+	 * in the outgoing RTP stream. */
+	LOGPLCGT(lchan, g_time, DRTP, LOGL_DEBUG,
+		 "Skipping RTP frame with lost payload\n");
+	if (lchan->abis_ip.osmux.use)
+		lchan_osmux_skipped_frame(lchan, fn_ms_adj(fn, lchan));
+	else if (lchan->abis_ip.rtp_socket)
+		osmo_rtp_skipped_frame(lchan->abis_ip.rtp_socket, fn_ms_adj(fn, lchan));
+	lchan->rtp_tx_marker = true;
+}
+
 /* TCH received from bts model */
 static int l1sap_tch_ind(struct gsm_bts_trx *trx, struct osmo_phsap_prim *l1sap,
 	struct ph_tch_param *tch_ind)
@@ -1580,12 +1942,12 @@ static int l1sap_tch_ind(struct gsm_bts_trx *trx, struct osmo_phsap_prim *l1sap,
 
 	gsm_fn2gsmtime(&g_time, fn);
 
-	LOGPGT(DL1P, LOGL_DEBUG, &g_time, "Rx TCH.ind chan_nr=%s\n", rsl_chan_nr_str(chan_nr));
-
 	lchan = get_active_lchan_by_chan_nr(trx, chan_nr);
 	if (!lchan) {
 		LOGPGT(DL1P, LOGL_ERROR, &g_time, "No lchan for TCH.ind (chan_nr=%s)\n", rsl_chan_nr_str(chan_nr));
 		return 0;
+	} else {
+		LOGPLCGT(lchan, &g_time, DL1P, LOGL_DEBUG, "Rx TCH.ind\n");
 	}
 
 	/* The ph_tch_param contained in the l1sap primitive may contain
@@ -1601,13 +1963,24 @@ static int l1sap_tch_ind(struct gsm_bts_trx *trx, struct osmo_phsap_prim *l1sap,
 	 * available is expected as empty payload. We also check if quality is
 	 * good enough. */
 	if (msg->len && tch_ind->lqual_cb >= bts->min_qual_norm) {
+		/* feed the good frame to the ECU, if we are applying one */
+		if (lchan->ecu_state)
+			osmo_ecu_frame_in(lchan->ecu_state, false, msg->data, msg->len);
 		/* hand msg to RTP code for transmission */
-		if (lchan->abis_ip.osmux.use) {
-			lchan_osmux_send_frame(lchan, msg->data, msg->len,
-					       fn_ms_adj(fn, lchan), lchan->rtp_tx_marker);
-		} else if (lchan->abis_ip.rtp_socket) {
-			osmo_rtp_send_frame_ext(lchan->abis_ip.rtp_socket,
-				msg->data, msg->len, fn_ms_adj(fn, lchan), lchan->rtp_tx_marker);
+		switch (lchan->rsl_cmode) {
+		case RSL_CMOD_SPD_SPEECH:
+			if (bts->emit_hr_rfc5993 && lchan->type == GSM_LCHAN_TCH_H &&
+			    lchan->tch_mode == GSM48_CMODE_SPEECH_V1)
+				send_rtp_rfc5993(lchan, fn, msg);
+			else
+				send_ul_rtp_packet_speech(lchan, fn, msg->data, msg->len);
+			break;
+		case RSL_CMOD_SPD_DATA:
+			send_ul_rtp_packet_data(lchan, fn, msg->data, msg->len);
+			break;
+		case RSL_CMOD_SPD_SIGN:
+		default: /* shall not happen */
+			OSMO_ASSERT(0);
 		}
 		/* if loopback is enabled, also queue received RTP data */
 		if (lchan->loopback) {
@@ -1616,16 +1989,8 @@ static int l1sap_tch_ind(struct gsm_bts_trx *trx, struct osmo_phsap_prim *l1sap,
 			/* Return 1 to signal that we're still using msg and it should not be freed */
 			return 1;
 		}
-		/* Only clear the marker bit once we have sent a RTP packet with it */
-		lchan->rtp_tx_marker = false;
 	} else {
-		DEBUGPGT(DRTP, &g_time, "Skipping RTP frame with lost payload (chan_nr=0x%02x)\n",
-			 chan_nr);
-		if (lchan->abis_ip.osmux.use)
-			lchan_osmux_skipped_frame(lchan, fn_ms_adj(fn, lchan));
-		else if (lchan->abis_ip.rtp_socket)
-			osmo_rtp_skipped_frame(lchan->abis_ip.rtp_socket, fn_ms_adj(fn, lchan));
-		lchan->rtp_tx_marker = true;
+		tch_ul_bfi_handler(lchan, &g_time, msg);
 	}
 
 	lchan->tch.last_fn = fn;
@@ -1670,17 +2035,26 @@ static bool rach_pass_filter(struct ph_rach_ind_param *rach_ind, struct gsm_bts
 	return true;
 }
 
-/* Special case where handover RACH is detected */
-static int l1sap_handover_rach(struct gsm_bts_trx *trx, struct ph_rach_ind_param *rach_ind)
+/* Special case where RACH on DCCH uplink is detected */
+static int l1sap_dcch_rach(struct gsm_bts_trx *trx, struct ph_rach_ind_param *rach_ind)
 {
+	struct gsm_lchan *lchan;
+
 	/* Filter out noise / interference / ghosts */
-	if (!rach_pass_filter(rach_ind, trx->bts, "handover")) {
+	if (!rach_pass_filter(rach_ind, trx->bts, "DCCH")) {
 		rate_ctr_inc2(trx->bts->ctrs, BTS_CTR_RACH_DROP);
 		return 0;
 	}
 
-	handover_rach(get_lchan_by_chan_nr(trx, rach_ind->chan_nr),
-		rach_ind->ra, rach_ind->acc_delay);
+	lchan = get_lchan_by_chan_nr(trx, rach_ind->chan_nr);
+	/* Differentiate + dispatch hand-over and VGCS RACH */
+	if (rsl_chan_rt_is_asci(lchan->rsl_chan_rt)) {
+		rate_ctr_inc2(trx->bts->ctrs, BTS_CTR_RACH_VGCS);
+		vgcs_rach(lchan, rach_ind->ra, rach_ind->acc_delay, rach_ind->fn);
+	} else {
+		rate_ctr_inc2(trx->bts->ctrs, BTS_CTR_RACH_HO);
+		handover_rach(lchan, rach_ind->ra, rach_ind->acc_delay);
+	}
 
 	/* must return 0, so in case of msg at l1sap, it will be freed */
 	return 0;
@@ -1731,8 +2105,7 @@ static int l1sap_ph_rach_ind(struct gsm_bts_trx *trx,
 		/* TODO: do we need to count Access Bursts on PDCH? */
 		return l1sap_pdch_rach(trx, rach_ind);
 	default:
-		rate_ctr_inc2(trx->bts->ctrs, BTS_CTR_RACH_HO);
-		return l1sap_handover_rach(trx, rach_ind);
+		return l1sap_dcch_rach(trx, rach_ind);
 	}
 
 	rate_ctr_inc2(trx->bts->ctrs, BTS_CTR_RACH_RCVD);
@@ -1882,16 +2255,44 @@ void l1sap_rtp_rx_cb(struct osmo_rtp_socket *rs, const uint8_t *rtp_pl,
 {
 	struct gsm_lchan *lchan = rs->priv;
 	struct msgb *msg;
+	bool rfc5993_sid = false;
 
 	/* if we're in loopback mode, we don't accept frames from the
 	 * RTP socket anymore */
 	if (lchan->loopback)
 		return;
 
-	msg = l1sap_msgb_alloc(rtp_pl_len);
+	/* initial preen */
+	switch (rtp_payload_input_preen(lchan, rtp_pl, rtp_pl_len, &rfc5993_sid)) {
+	case PL_DECISION_DROP:
+		return;
+	case PL_DECISION_ACCEPT:
+		break;
+	case PL_DECISION_STRIP_HDR_OCTET:
+		rtp_pl++;
+		rtp_pl_len--;
+		break;
+	default:
+		OSMO_ASSERT(0);
+	}
+
+	msg = l1sap_msgb_alloc(512);
 	if (!msg)
 		return;
-	memcpy(msgb_put(msg, rtp_pl_len), rtp_pl, rtp_pl_len);
+
+	if (lchan->rsl_cmode == RSL_CMOD_SPD_DATA) {
+		int rc = csd_v110_rtp_decode(lchan, msg->tail,
+					     rtp_pl, rtp_pl_len);
+		if (rc > 0) {
+			msgb_put(msg, rc);
+		} else {
+			msgb_free(msg);
+			return;
+		}
+	} else {
+		memcpy(msgb_put(msg, rtp_pl_len), rtp_pl, rtp_pl_len);
+	}
+
 	msgb_pull(msg, sizeof(struct osmo_phsap_prim));
 
 	/* Store RTP header Marker bit in control buffer */
@@ -1900,6 +2301,8 @@ void l1sap_rtp_rx_cb(struct osmo_rtp_socket *rs, const uint8_t *rtp_pl,
 	rtpmsg_seq(msg) = seq_number;
 	/* Store RTP header Timestamp in control buffer */
 	rtpmsg_ts(msg) = timestamp;
+	/* Store RFC 5993 SID flag likewise */
+	rtpmsg_is_rfc5993_sid(msg) = rfc5993_sid;
 
 	/* make sure the queue doesn't get too long */
 	lchan_dl_tch_queue_enqueue(lchan, msg, 1);
@@ -1933,12 +2336,19 @@ int l1sap_chan_act(struct gsm_bts_trx *trx, uint8_t chan_nr)
 
 	LOGPLCHAN(lchan, DL1C, LOGL_INFO, "Activating channel %s\n", rsl_chan_nr_str(chan_nr));
 
-	lchan->s = trx->bts->radio_link_timeout.current;
+	radio_link_timeout_reset(lchan);
 
 	rc = l1sap_chan_act_dact_modify(trx, chan_nr, PRIM_INFO_ACTIVATE, 0);
 	if (rc)
 		return -RSL_ERR_EQUIPMENT_FAIL;
 
+	/* Is it TCH?  If it is, attempt to allocate an Error Concealment Unit
+	 * instance, if available, unless it is disabled by vty config. */
+	if (lchan_is_tch(lchan) && trx->bts->use_ul_ecu)
+		lchan->ecu_state = osmo_ecu_init(trx, lchan2ecu_codec(lchan));
+	else
+		lchan->ecu_state = NULL;
+
 	/* Init DTX DL FSM if necessary */
 	if (trx->bts->dtxd && lchan_is_tch(lchan)) {
 		lchan->tch.dtx.dl_amr_fsm = osmo_fsm_inst_alloc(&dtx_dl_amr_fsm,
@@ -1979,6 +2389,12 @@ int l1sap_chan_rel(struct gsm_bts_trx *trx, uint8_t chan_nr)
 		lchan->tch.dtx.dl_amr_fsm = NULL;
 	}
 
+	/* clear ECU state (if any) */
+	if (lchan->ecu_state) {
+		osmo_ecu_destroy(lchan->ecu_state);
+		lchan->ecu_state = NULL;
+	}
+
 	return l1sap_chan_act_dact_modify(trx, chan_nr, PRIM_INFO_DEACTIVATE,
 		0);
 }
@@ -2001,5 +2417,18 @@ int l1sap_chan_modify(struct gsm_bts_trx *trx, uint8_t chan_nr)
 	LOGPLCHAN(lchan, DL1C, LOGL_INFO, "Modifying channel %s\n",
 		rsl_chan_nr_str(chan_nr));
 
+	/* Is it TCH?  If it is and we are applying internal uplink ECUs,
+	 * the new channel mode calls for a different ECU.  Any changes
+	 * in vty config (enabling or disabling this ECU application)
+	 * will also take effect upon channel modification. */
+	if (lchan_is_tch(lchan)) {
+		if (lchan->ecu_state)
+			osmo_ecu_destroy(lchan->ecu_state);
+		if (trx->bts->use_ul_ecu)
+			lchan->ecu_state = osmo_ecu_init(trx, lchan2ecu_codec(lchan));
+		else
+			lchan->ecu_state = NULL;
+	}
+
 	return l1sap_chan_act_dact_modify(trx, chan_nr, PRIM_INFO_MODIFY, 0);
 }
